import boto3
import re

import psychic_disco
console = boto3.client('apigateway')


def fetch_api_by_name(api_name):
    """ Fetch an api record by its name """
    api_records = console.get_rest_apis()['items']
    matches = filter(lambda x: x['name'] == api_name, api_records)
    if not matches:
        return None
    return matches[0]


def fetch_resources_by_api(api_id):
    """ Fetch all resources for an apigateway rest api """
    return console.get_resources(restApiId=api_id)['items']


def fetch_resource(api_id, resource_id):
    """ Fetch extra metadata for this particular resource """
    return console.get_resource(restApiId=api_id, resourceId=resource_id)


def fetch_method(api_id, resource_id, verb):
    """ Fetch extra metadata for this particular method """
    return console.get_method(
            restApiId=api_id,
            resourceId=resource_id,
            httpMethod=verb)


def lambda_arn_from_method_integration_arn(int_arn):
    pattern = '.*functions/(.*)/invocations'
    matches = re.match(pattern, int_arn)
    return matches.group(1)


def lambda_name_from_lambda_arn(lambda_arn):
    return lambda_arn.split(':')[-1]


def fetch_root_resource_for_api(api_id):
    records = console.get_resources(restApiId=api_id)['items']
    for resource in records:
        if resource['path'] == '/':
            return resource


class RestApi(object):
    apis_by_name = dict()

    def __init__(self, name):
        self.name = name
        self.aws_id = None
        self.resources = dict()

    @classmethod
    def find(cls, api_name):
        if api_name in cls.apis_by_name:
            return cls.apis_by_name[api_name]
        api = cls(api_name)
        api._fetch_from_aws()
        if api.exists_in_aws:
            api._fetch_resources()
        cls.apis_by_name[api_name] = api
        return api

    @property
    def root_resource(self):
        root = self.declare_resource('/')
        if not root.exists_in_aws:
            record = fetch_root_resource_for_api(self.aws_id)
            root.aws_id = record['id']
        return root

    def _fetch_from_aws(self):
        record = fetch_api_by_name(self.name)
        if record:
            self.aws_id = record['id']

    @property
    def exists_in_aws(self):
        return (self.aws_id is not None)

    def _fetch_resources(self):
        for record in fetch_resources_by_api(self.aws_id):
            path = record['path']
            if path not in self.resources:
                self.resources[path] = RestResource(path)
            resource = self.resources[path]
            resource.aws_id = record['id']
            resource.api = self
            resource.fetch_from_aws()

    def deploy(self):
        print "Deploying API: %s" % self.name
        self._fetch_from_aws()
        if not self.exists_in_aws:
            console.create_rest_api(name=self.name)
            self._fetch_from_aws()
        self.root_resource.deploy()  # It's special
        for path in sorted(self.resources.keys()):
            resource = self.resources[path]
            resource.deploy()
        console.create_deployment(
                restApiId=self.aws_id,
                stageName='master')

    def declare_route(self, verb, path, lambda_function):
        r = self.declare_resource(path)
        r.declare_route(verb, lambda_function)

    def declare_resource(self, path):
        if path not in self.resources:
            r = RestResource(path)
            r.api = self
            self.resources[path] = r
        return self.resources[path]


class RestResource(object):
    def __init__(self, path):
        self.path = path
        self.aws_id = None
        self.api = None
        self.methods = dict()

    @property
    def exists_in_aws(self):
        return (self.aws_id is not None and self.api is not None)

    def fetch_from_aws(self):
        if not self.exists_in_aws:
            raise Exception("Not enough info to query AWS for Rest Resource")
        record = fetch_resource(self.api.aws_id, self.aws_id)
        if 'resourceMethods' in record:
            for verb in record['resourceMethods']:
                if verb not in self.methods:
                    self.methods[verb] = RestMethod(verb)
                method = self.methods[verb]
                method.resource = self
                method.api = self.api
                method.fetch_from_aws()

    @property
    def parent(self):
        if self.path == '/':
            return None
        else:
            parent_path = '/'.join(self.path.split('/')[0:-1]) or '/'
            return self.api.declare_resource(parent_path)

    @property
    def path_part(self):
        if self.parent:
            return self.path.replace(self.parent.path + "/", '', 1)
        else:
            return self.path

    def deploy(self):
        print "\tDeploying Resource: %s" % self.path
        if not self.exists_in_aws:
            self.parent.deploy()
            response = console.create_resource(
                    restApiId=self.api.aws_id,
                    parentId=self.parent.aws_id,
                    pathPart=self.path_part)
            self.aws_id = response['id']
        for verb in self.methods:
            method = self.methods[verb]
            method.deploy()

    def declare_method(self, verb):
        if verb not in self.methods:
            m = RestMethod(verb)
            m.resource = self
            m.api = self.api
            self.methods[verb] = m
        return self.methods[verb]

    def declare_route(self, verb, lambda_function):
        m = self.declare_method(verb)
        m.declare_route(lambda_function)


class RestMethod(object):
    def __init__(self, verb):
        self.verb = verb
        self.resource = None
        self.api = None
        self.lambda_name = None
        self.lambda_function = None

    def declare_route(self, lf):
        self.lambda_function = lf
        self.lambda_name = lf.proper_name

    def fetch_from_aws(self):
        if (self.resource is None or self.api is None):
            msg = """I cannot possibly question AWS
            without knowing my resource and api ids"""
            raise Exception(msg)
        record = fetch_method(self.api.aws_id, self.resource.aws_id, self.verb)
        arn = record['methodIntegration']['uri']
        lambda_arn = lambda_arn_from_method_integration_arn(arn)
        self.lambda_name = lambda_name_from_lambda_arn(lambda_arn)

    @property
    def exists_in_aws(self):
        try:
            console.get_method(
                    restApiId=self.api.aws_id,
                    resourceId=self.resource.aws_id,
                    httpMethod=self.verb)
            return True
        except:
            return False

    @property
    def integration_uri(self):
        pattern = "arn:aws:apigateway:%s:lambda:path/2015-03-31/functions/%s/invocations"  # noqa
        region = psychic_disco.config.default_aws_region
        lambda_arn = self.lambda_function.arn
        return pattern % (region, lambda_arn)

    def deploy(self):
        print "\t\tDeploying Method: %s" % self.verb
        should_put_integration = False
        if not self.exists_in_aws:
            should_put_integration = True
            console.put_method(
                    restApiId=self.api.aws_id,
                    resourceId=self.resource.aws_id,
                    httpMethod=self.verb,
                    authorizationType='NONE')
        if self.lambda_function:
            self.lambda_function.deploy()
        if should_put_integration:
            console.put_integration(
                    restApiId=self.api.aws_id,
                    resourceId=self.resource.aws_id,
                    httpMethod=self.verb,
                    type='AWS',
                    integrationHttpMethod=self.verb,
                    uri=self.integration_uri)



import unittest
import psychic_disco
import shutil
import os

zip_file_path = '.psychic_disco/deployment-package.zip'


class TestDeploymentPackage(unittest.TestCase):
    def test_default_is_same(self):
        dp = psychic_disco.deployment_package.default()
        dp2 = psychic_disco.deployment_package.default()
        self.assertEqual(dp, dp2)

    def test_localpath(self):
        dp = psychic_disco.deployment_package.default()
        self.assertEqual(zip_file_path, dp.localpath)


class TestDeploymentPackageOperations(unittest.TestCase):
    def setUp(self):
        self.dp = psychic_disco.deployment_package.default()
        shutil.rmtree("poop", True)
        os.mkdir("poop")
        os.chdir("poop")

    def tearDown(self):
        os.chdir("..")
        shutil.rmtree("poop", True)

    def test_setup_virtualenv(self):
        with open("requirements.txt", "w") as f:
            f.write("Jinja2==2.8\npsychic_disco==0.6.0\n")
        installed_packages = self.dp.setup_virtualenv("venv")
        self.assertIn(('Jinja2', '2.8'), installed_packages)
        self.assertIn(('psychic-disco', '0.6.0'), installed_packages)

    def test_bundle(self):
        with open("poop1.py", "w") as f:
            f.write("print 'hello world'\n")
        self.dp.bundle()
        self.assertTrue(os.path.exists(zip_file_path))


if __name__ == "__main__":
    unittest.main()



import boto3
import deployment_package
import config

console = boto3.client('lambda')
_declared_functions = []


def installed_functions():
    return [f['FunctionName'] for f in console.list_functions()['Functions']]


def declared_functions():
    return _declared_functions


class LambdaFunction(object):
    def __init__(self, func_name, module_name):
        global _declared_functions
        self.full_name = "%s.%s" % (module_name, func_name)
        _declared_functions.append(self)

    @property
    def arn(self):
        aws_record = console.get_function(FunctionName=self.proper_name)
        return aws_record['Configuration']['FunctionArn']

    @property
    def proper_name(self):
        return self.full_name.replace('.', '-')

    def update(self, package):
        print("Updating %s" % self.proper_name)
        console.update_function_code(
                FunctionName=self.proper_name,
                S3Bucket=package.bucket,
                S3Key=package.key)

    def install(self, package):
        print("Installing %s" % self.proper_name)
        console.create_function(
                FunctionName=self.proper_name,
                Runtime="python2.7",
                Role=config.default_iam_role,
                Handler=self.full_name,
                Code={"S3Bucket": package.bucket, "S3Key": package.key})

    def deploy(self):
        package = deployment_package.default()
        package.deploy()
        if self.proper_name in installed_functions():
            self.update(package)
        else:
            self.install(package)

    def __str__(self):
        return self.proper_name



import unittest
import psychic_disco.api_gateway_config as agc
import json
import os


class FakeApiGatewayConsole(object):
    def __init__(self):
        self.apis = dict()
        self.basepath = os.path.dirname(os.path.abspath(__file__))

    def _open_json(self, filename):
        path = os.path.join(self.basepath, filename)
        with open(path) as f:
            return json.loads(f.read())

    def get_rest_apis(self):
        return self._open_json("get_rest_apis.json")

    def get_resources(self, **kwargs):
        return self._open_json("get-resources.json")

    def get_resource(self, **kwargs):
        return self._open_json("get-resource.json")

    def get_method(self, **kwargs):
        return self._open_json("get-method.json")


class TestApiGatewayConfig(unittest.TestCase):
    def setUp(self):
        agc.console = FakeApiGatewayConsole()

    def test_fetch_api_by_name(self):
        agc.console.apis["good_api"] = agc.RestApi.find("good_api")
        actual = agc.fetch_api_by_name("dummy-manager")
        self.assertEqual("123apiid", actual['id'])

    def test_fetch_wrong_api_name(self):
        actual = agc.fetch_api_by_name("an api has no name")
        self.assertIsNone(actual)

    def test_fetch_resources_by_api(self):
        actual = agc.fetch_resources_by_api("123apiid")
        self.assertEqual(2, len(actual))
        self.assertEqual("/", actual[0]['path'])
        self.assertEqual("/dummies", actual[1]['path'])

    def test_fetch_resource(self):
        actual = agc.fetch_resource("123apiid", "123resourceid")
        self.assertEqual({}, actual['resourceMethods']['post'])

    def test_fetch_method(self):
        actual = agc.fetch_method("123apiid", "123resourceid", "post")
        uri = actual['methodIntegration']['uri']
        expected_ending = "psychic_disco-dummy_lambda_function/invocations"
        self.assertTrue(uri.endswith(expected_ending))

    def test_lambda_arn_from_method_integration_arn(self):
        method = agc.console._open_json("get-method.json")
        uri = method['methodIntegration']['uri']
        expected_arn = "arn:aws:lambda:us-east-1:123456789012:function:psychic_disco-dummy_lambda_function"  # noqa
        actual_arn = agc.lambda_arn_from_method_integration_arn(uri)
        self.assertEqual(expected_arn, actual_arn)

    def test_lambda_name_from_lambda_arn(self):
        arn = "arn:aws:lambda:us-east-1:123456789012:function:psychic_disco-dummy_lambda_function"  # noqa
        expected_name = "psychic_disco-dummy_lambda_function"
        actual_name = agc.lambda_name_from_lambda_arn(arn)
        self.assertEqual(expected_name, actual_name)

    def test_rest_api_find(self):
        api = agc.RestApi.find("dummy-manager")
        self.assertEqual("123apiid", api.aws_id)
        self.assertTrue(api.exists_in_aws)
        self.assertEqual(2, len(api.resources))
        self.assertIn('/', api.resources)
        self.assertIn('/dummies', api.resources)

    def test_rest_resource_not_exists(self):
        resource = agc.RestResource('/dummies')
        self.assertFalse(resource.exists_in_aws)

    def test_rest_resource_exists(self):
        api = agc.RestApi.find("dummy-manager")
        resource = agc.RestResource('/dummies')
        resource.aws_id = "123resourceid"
        resource.api = api
        self.assertTrue(resource.exists_in_aws)

    def test_rest_resource_fetch_from_aws(self):
        api = agc.RestApi.find("dummy-manager")
        resource = agc.RestResource('/dummies')
        resource.aws_id = "123resourceid"
        resource.api = api

        self.assertEqual(0, len(resource.methods))
        resource.fetch_from_aws()
        self.assertEqual(1, len(resource.methods))

    def test_rest_resource_path_part(self):
        api = agc.RestApi.find("dummy-manager")
        resource = agc.RestResource('/dummies/by_name')
        resource.api = api
        self.assertEqual(resource.path_part, "by_name")

    def test_method_fetch_from_aws(self):
        method = agc.RestMethod('post')
        method.resource = agc.RestResource('/dummies')
        method.resource.aws_id = "123resourceid"
        method.resource.api = agc.RestApi.find("dummy-manager")
        method.api = method.resource.api

        self.assertIsNone(method.lambda_name)
        method.fetch_from_aws()
        self.assertEqual(
                "psychic_disco-dummy_lambda_function",
                method.lambda_name)


if __name__ == "__main__":
    unittest.main()
